/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cnode.hpp
 *		- abstract class for declaration 
 *
 *
 ******************************************************************************
 */

#ifndef CNODE_HPP
#define CNODE_HPP

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <math.h>
#include <list>
#include <set>
#include "glue.h"
#include "csymbol.h"
#include "cdecl.h"
#include "cvector.h"
#include "cobstack.h"
#include "cattr.h"

class CGenvar;
class CParam;
class CFref;
class CVar;
class CNet;
class CPort;
class CPortDir;
class CInstance;
class CFunction;
class CModule;
class CGate;
class CEvent;
class CAttr;
class CBlock;
class CSpecify;
class CNode;
class CTypedef;
class CEnum;
typedef CBlock CScope;
inline char* d2s( double d, CObstack* stack );

/**
 * Edge values
 */
enum Edge_t {
	eEDGE01 = 0x1,
	eEDGE10 = 0x2,
	eEDGE0x = 0x4,
	eEDGEx1 = 0x8,
	eEDGE1x = 0x10,
	eEDGEx0 = 0x20,
};
	
/**
 * Timing mode values
 */
enum DelayMode_t {
	eMIN_DELAY,
	eTYP_DELAY,
	eMAX_DELAY
};

/**
 * Strength values
 */
enum Strength_t {
	eUNDEFINED = 0,
	eSUPPLY,
	eSTRONG,
	ePULL,
	eLARGE,
	eWEAK,
	eMEDIUM,
	eSMALL,
	eHIGHZ,
};

/**
 * Pair of strengths
 */
struct StrengthPair_t {
	Strength_t	s0;
	Strength_t	s1;
};

#define DEFINE_ENUM
#include "cnode_def.h"
#undef DEFINE_ENUM


/**
 * Smart pointer for CNode class
 * Creates safe references to CNode arguments 
 * Supports assignment, dereference, and 
 * equality operations
 */
template<class T>
class CNode_sp {
	T* ptr;	/// Pointer to untyped argument
public:
	CNode_sp( void** np ) { ptr = (T*)np; }
	T operator=( T n ) { *ptr = n; return n; }
	T operator->() { return *ptr; }
	operator T() { return *ptr; }
	int operator==( T v ) { return *ptr == v; }
	int operator!=( T v ) { return *ptr != v; }
	int operator==( CNode_sp<T> p ) { return *ptr == *p.ptr; }
	int operator!=( CNode_sp<T> p ) { return *ptr != *p.ptr; }
};

/**
 * Helper class for building tail recursive binary CNode trees
 * Used by parser.
 */
struct CNode_pr {
	CNode* head;
	CNode* tail;
public:
	CNode* operator=( CNode* n ) { head = n; tail = n; return n; }
	CNode* operator->() { return head; }
	operator CNode*() { return head; }
	int operator==( CNode* v ) { return head == v; }
	int operator!=( CNode* v ) { return head != v; }
	int operator==( CNode_pr p ) { return head == p.head; }
	int operator!=( CNode_pr p ) { return head != p.head; }
	friend CNode_pr	cLINK( CNode_pr pr1, CNode* n2 );
};

/**
 * Helper class for building tail recursive binary CNode trees
 * Used by parser.
 */
struct CNode_triplet {
	CNode* first;
	CNode* second;
	CNode* third;
};


/**
 * Primary data structure representing parse tree nodes.
 * Each parse tree node is represented as a forest of 
 * binary CNode objects. Nodes are always balenced with
 * left to right construction.
 *
 * CNode uses the CObstack memory allocator (via CObject).
 * This allows memory to be allocated in large slabs and 
 * released enmasse.
 *
 * Each parse node represents an operation, represented by NodeOp_t. 
 * Additionally expression nodes may also be decorated with
 * file/line info, verilog attributes, and temporary tags.
 * All decorations are only valid on the CNode at the root of 
 * the parse nodes CNode graph.
 *
 * The operands of a parse tree node are referenced using the 
 * method Arg. A description of the parse tree nodes and 
 * there arguments can be found here: \ref parse_tree_node .
 */ 
class CNode : public CObject
{
private:
	static list<CObstack*> stackList; ///< stack of heaps in use
	static CObstack		evalHeap; ///< temporary memory heap used for expression evaluation
	static INT32		evalCount; ///< track the number of recursive uses of the evaluation heap
	static CObstack* stack;	    ///< pointer to current allocator
        static map<CNode*,int> labelCache;  ///< used to accellerate LabelBits.
	static int labelCacheEnabled;   ///< non-zero if label cache is enabled.
	NodeOp_t	op;	    ///< Node's operation type
	void*		left;	    ///< left child
	void*		right;	    ///< right child
	Coord_t		loc;	    ///< embedded storage for file location info
	Coord_t		*locp;	    ///< pointer to location info
	CNode*		attributes; ///< verilog attributes
	/*
 	 * These decorations are temporary and used
 	 * by the expression evaluation routines
 	 */
	NodeType_t	type;	    ///< transient: Expression type
	INT32		width;	    ///< transient: Expression width in bits
	int		fixedWidth; ///< transient: True if width is fixed
private:
	int	LabelBits( int supressErrorMessages = FALSE );
	CNode*	FixBits( INT32 newWidth, NodeType_t newType );
	void	_EvalVector( CVector& v );
	double	_EvalReal( void );
	void FixedWidth( int v ) { fixedWidth = v; }
	int  FixedWidth() { return fixedWidth; }
	int  ConditionalWiden();
	int  WidthFixed();
	unsigned  NodeMask();
	CNode*	GetNLeft( void ) { return (CNode*)left; }
	CNode*	GetNRight( void ) { return (CNode*)right; }
	static void _LabelBits( CNode* n, void* arg );
public:
	/**
 	 * Gets pointer to current heap allocator
 	 * \return pointer to current heap allocator
 	 */
	static CObstack* CurrentHeap() { return stack; }
	/**
         * Use evaluation stack. Track recursive call so that eval heap
         * is only freed when all uses are complete. Note that nodes
         * on the eval stack should *NEVER* point to nodes that are not.
         * If you do the labelCache will cause stale information to be
         * used with bizare results. Usually this means that subtrees
         * attached to evalStack allocated node should be cloned on to
         * the evalStack.
         */
	static void UseEvalStack( void ) {
		stackList.push_front( stack );
		evalCount++;
		stack = &evalHeap;
	}
	/**
 	 * Set heap to a specific heap. Save previous heap in stack
 	 * so it can be restored
 	 */
	static void SetBuildStack( CObstack* aStack ) {
		MASSERT( evalCount == 0 );
		stackList.push_front( stack );
		stack = aStack;
	}
	/**
 	 * Restore previous heap
 	 */
	static void ResetBuildStack( void ) {
		if( stack == &evalHeap ) {
			evalCount--;
			if( evalCount == 0 ) {
				evalHeap.Free( NULL );
			}
		}
		if( stackList.empty() ) {
			stack = NULL;
		} else {
			stack = *stackList.begin();
			stackList.pop_front();
		}
	}
	/**
 	 * Enable cache of labeled nodes to be tracked.
 	 * This greatly speeds up operations which utilize
 	 * expression width and type information. However it
 	 * should only be enabled when you know you are not changing
 	 * the tree structures in such a way that the type and width
 	 * properties change. This cache should always be disable and
 	 * cleared when your done with it.
 	 */
	static void EnableLabelCache()
	{
	    labelCacheEnabled = 1;
	}
	/**
 	 * Disable caching of label info (width and type) and clear
 	 * all accumulated data.
 	 */
	static void DisableAndClearLabelCache()
	{
	    labelCacheEnabled = 0;
	    labelCache.erase( labelCache.begin(), labelCache.end() );
	}
	/**
 	 * Constructor for parse node.
 	 * Note this routine will create a linked set of CNode for nodes
 	 * with more than 2 operands.
 	 * \param aLoc is file info for the parse node.
 	 * \param aOp is the parse node operation.
 	 */
	CNode( Coord_t* aLoc, NodeOp_t aOp ); 
	/**
 	 * Get node's file coordinates
 	 * \return pointer to node's file coordinates
 	 */
	Coord_t* GetCoord() { return locp; }
	/**
 	 * Return node's operation type
 	 * \return node's operation type
 	 */
	NodeOp_t GetOp() { return op; }
	/**
 	 * Set node's operation type.
 	 * Asserts if number arguments in new type are different 
 	 * than old.
 	 * \param aOp is the new operation type.
 	 */
	void SetOp( NodeOp_t aOp ) { 
		int oldCount = ArgCount();
		op = aOp; 
		MASSERT( oldCount == ArgCount() );
	}
	/**
 	 * Calculate hash of tree.
 	 * Equivilent(n1,n2) => Hash(n1)==Hash(n2) 
 	 * \returns tree hash code.
 	 */
	unsigned Hash();
	/**
 	 * Get a node's operand.
 	 * \param index is a the operand number (zero based).
 	 * \return type safe operand "number" of node
	 */
	template<class T> CNode_sp<T> Arg( int index );
	/**
 	 * Get the number of operands for the node.
	 * \return number of operands.
	 */
	int	ArgCount( void );
	/**
 	 * Replicate tree. 
 	 * Leaf nodes are not replicated.
 	 * \param heap heap to use for allocation, defaults to
 	 *        current node heap.
 	 * \return shallow copy of tree.
 	 */
	CNode*	Clone( CObstack* heap = stack );
	/**
 	 * Get the precedence of the operator represented by the node
 	 * \return precedence of the operator
 	 */
	int     Precedence();
	/**
 	 * Walk tree invoking callback on each node after children have 
 	 * been visited.
	 * \param callback function to be called for each node
	 * \param data opaque argument passed to callback function
	 */
	void	PostVisit1( void (*callback)(CNode*,void*), void* data );
	/**
 	 * Walk tree invoking callback on each node after children have 
 	 * been visited. callback may modify tree and return a modified
 	 * subtree.
	 * \param callback function to be called for each node
	 * \param data opaque argument passed to callback function
	 * \return modified child subtree.
	 */
	CNode*	PostSubVisit1( CNode* (*callback)(CNode*,void*), 
							void* data );
	/**
 	 * Walk tree invoking callback on each node before children have 
 	 * been visited. callback can terminate transversal early.
	 * callback returns 0 to terminate transversal early.
	 * \param callback function to be called for each node
	 * \param data opaque argument passed to callback function
	 */
	void	PreVisit1( int (*callback)(CNode*,void*), void* data );
	/**
 	 * Create simplified expression tree with given width and type.
 	 * Evaluates non-volatile constant expressions.
	 * \param newWidth width of resulting expression.
	 * \param newType type of resulting expression.
	 * \return simplified expression if expression can be simplified,
	 * otherwise original expression is returned.
	 */
	CNode*	Simplify( INT32 newWidth, NodeType_t newType );
	/**
 	 * Checks expression tree to see if expression can result in an
 	 * X or Z. 
 	 * \param integerIsNonX non-zero specifies that integers should 
 	 *        considered to never take on X or Z values.
 	 * \param exclude specifies an attribute name that indicates
 	 *        the declaration will never be X or Z.
 	 * \return non-zero indicates the expression will never be X or Z.
	 */
	int	IsNonX( int integerIsNonX = 0, char* exclude = NULL );
	/**
 	 * Checks expression tree to see if it is constant.
 	 * i.e. Dependent on constants and parameters only.
 	 * \return non-zero indicates expression is constant.
 	 */
	int	IsConstant();
	/**
 	 * Checks to see if expression tree can be evaluated.
 	 * Expressions may not be able to be evaluated for a variety 
 	 * of reasons:
 	 * - Contains a construct the expression evaluator doesn't support.
 	 * - Contains a reference to a variable.
 	 * - Constains an external reference.
 	 * \return non-zero if expression can be evaluated.
 	 */
	int	IsEvaluateable();
	/**
 	 * Checks to see if expression tree is volatile. i.e.
 	 * Dependent upon parameters and variables.
 	 * \return non-zero if expression is volatile.
 	 */
	int	IsVolatile( void );
	/**
 	 * Evaluates expression tree and returns value as a 32 bit integer.
 	 * Expression tree must be evaluateable otherwise results are undefined.	 * \return expression value evaluated as an integer.
	 */
	INT32	EvalINT32();
	/**
 	 * Evaluates expression tree evaluated in unconstrainted context. 
 	 * Result is returned as a vector.
 	 * Expression tree must be evaluateable otherwise results are undefined.
 	 * \param v reference to vector to store result
	 */
	void 	EvalVector( CVector& v );
	/**
 	 * Evaluates expression tree evaluated in the specified context. 
 	 * Result is returned as a vector.
 	 * Expression tree must be evaluateable otherwise results are undefined.
 	 * \param v reference to vector to store result
 	 * \param newWidth width of evaluating context.
 	 * \param newType  type of evaluating context.
	 */
	void 	EvalVector( CVector& v, INT32 newWidth, NodeType_t newType );
	/**
 	 * Evaluates expression tree evaluated in a real context.
 	 * Expression tree must be evaluateable otherwise results are undefined.
 	 * \return expression value evaluated as a real.
	 */
	double  EvalReal( void );
	/**
 	 * Print a compact representation of the parse tree
 	 * \param f file descriptor to print output.
 	 */
	void	Dump( FILE* f );
	/**
 	 * Evaluates if expression width is constant. i.e. Depends upon
 	 * parameters and constants only.
 	 * \return non-zero if expression width is constant.
 	 */
	int	IsWidthConstant( void );
	/**
 	 * Evaluates if expression width is volatile. i.e. Depends upon
 	 * parameters and variables.
 	 * \return non-zero if expression width is volatile.
 	 */
	int	IsWidthVolatile( void );
	/**
 	 * Evaluates if expression width can be evaluated. 
 	 * Widths may not be able to be evaluated for a variety 
 	 * of reasons:
 	 * - Contains a construct the expression evaluator doesn't support.
 	 * - Contains a reference to a variable.
 	 * - Constains an external reference.
 	 * \return non-zero if expression width is volatile.
 	 */
	int	IsWidthEvaluateable( void );
	/**
 	 * Create expression representing width of expression
 	 * \return expression representing width of expression
 	 */
	CNode*	GetWidthExp( void );
	/**
 	 * Evaluate width of expression. Width must be evaluateable,
 	 * otherwise the results are undefined.
 	 * \return width of expression.
 	 */
	INT32   GetWidth( void ) { LabelBits(TRUE); return width; }
	/**
 	 * Determine if expression is a 1 bit signed or unsigned value.
 	 * \return non-zero if expression is scalar.
 	 */
	int	IsScalar( void ) { LabelBits(TRUE); return width==1; }
	/**
 	 * Determine if expression is a multi-bit signed or unsigned value.
 	 * \return non-zero if expression is a vector.
 	 */
	int	IsVector( void ) { LabelBits(TRUE); return width>1; }
	/**
 	 * Determine if expression is real.
 	 * \return non-zero if expression is a real.
 	 */
	int	IsReal( void ) { LabelBits(TRUE); return type==eR; }
	/**
 	 * Get attributes attached to operation.
 	 * \return attribute parse tree. NULL if no attibutes are attached.
 	 */
	CNode*  GetAttributes() { return attributes; }
	/**
 	 * Attach attributes to operation.
 	 * \param attr attributes to attach to operation.
 	 */
	void	SetAttributes( CNode* attr ) { attributes = attr; }
	/**
 	 * Determine if node has the given attribute.
 	 * \param name name of attribute to search for.
 	 * \param n don't supply this argument.
 	 * \param init don't supply this argument.
 	 * \return non-zero if node has attribute
 	 */
        int   HasAttribute( const char* name, CNode* n=NULL, int init = 1 );
	/**
 	 * Get attribute attached to node with the given attribute.
 	 * \param name name of attribute to search for.
 	 * \param n don't supply this argument.
 	 * \param init don't supply this argument.
 	 * \return attribute matching name. NULL if none.
 	 */
        CAttr* GetAttribute( const char* name, CNode* n=NULL, int init = 1 );
	/**
 	 * Get node expression type.
 	 * \return expression type.
 	 */
	NodeType_t GetNodeType( void ) { LabelBits(TRUE); return type; }
};


/************************************************
 	Arg<CNode*>	
	- returns CNode smart pointer to arg by index
 ***********************************************/

template<class T> CNode_sp<T> CNode::Arg(int index) 
{
	switch( ArgCount() ) {
	case 1:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&left);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 2:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&left);		
		case 1:
			return CNode_sp<T>(&right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 3:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 4:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&GetNRight()->left);		
		case 3:
			return CNode_sp<T>(&GetNRight()->right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 5:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&GetNLeft()->right);		
		case 3:
			return CNode_sp<T>(&GetNRight()->left);		
		case 4:
			return CNode_sp<T>(&GetNRight()->right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 6:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->left);		
		case 3:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->right);		
		case 4:
			return CNode_sp<T>(&GetNRight()->left);		
		case 5:
			return CNode_sp<T>(&GetNRight()->right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 7:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->left);		
		case 3:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->right);		
		case 4:
			return CNode_sp<T>(&GetNRight()->GetNLeft()->left);		
		case 5:
			return CNode_sp<T>(&GetNRight()->GetNLeft()->right);		
		case 6:
			return CNode_sp<T>(&GetNRight()->right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
	case 8:
		switch( index ) {
		case 0:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->left);		
		case 1:
			return CNode_sp<T>(&GetNLeft()->GetNLeft()->right);		
		case 2:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->left);		
		case 3:
			return CNode_sp<T>(&GetNLeft()->GetNRight()->right);		
		case 4:
			return CNode_sp<T>(&GetNRight()->GetNLeft()->left);		
		case 5:
			return CNode_sp<T>(&GetNRight()->GetNLeft()->right);		
		case 6:
			return CNode_sp<T>(&GetNRight()->GetNRight()->left);		
		case 7:
			return CNode_sp<T>(&GetNRight()->GetNRight()->right);		
		default:
			MASSERT( FALSE );
			return NULL;
		} 
		
	default:
		MASSERT( FALSE );		
	}
}
int Equivalent( CNode* a, CNode* b );
/******************************************************
  real operation routines
 ******************************************************/

inline void Add( double* r, double* a, double* b )
{
	*r = *a + *b;
}

inline void Sub( double* r, double* a, double* b )
{
	*r = *a - *b;
}

inline void Mul( double* r, double* a, double* b )
{
	*r = *a * *b;
}

inline void Div( double* r, double* a, double* b )
{
	*r = *a / *b;
}

inline void Neg( double* r, double* a )
{
	*r = - *a; 
}

inline void Plus( double* r, double* a )
{
	*r = *a; 
}

inline void Pow( double* r, double* a, double* b )
{
	*r = pow(*a,*b);
}


/*****************************************************
 * Create stubs for illegal operations
 ****************************************************/
#define ILLEGAL_OP2(op) \
inline void op( double*, double*, double* )\
{ fatal( NULL, #op " is illegal for reals" ); }

#define ILLEGAL_OP1(op) \
inline void op( double*, double* )\
{ fatal( NULL, #op " is illegal for reals" ); }

ILLEGAL_OP2(Rsh);
ILLEGAL_OP2(Lsh);

ILLEGAL_OP2(Rep);
ILLEGAL_OP2(Mod);
ILLEGAL_OP2(And);
ILLEGAL_OP2(Xor);
ILLEGAL_OP2(Xnor);
ILLEGAL_OP2(Or);
ILLEGAL_OP2(Lor);
ILLEGAL_OP2(Land);
ILLEGAL_OP1(Com);
ILLEGAL_OP1(Rand);
ILLEGAL_OP1(Rnand);
ILLEGAL_OP1(Ror);
ILLEGAL_OP1(Rnor);
ILLEGAL_OP1(Rxor);
ILLEGAL_OP1(Rxnor);

#define DEFINE_CONSTRUCTOR
#include "cnode_def.h"
#undef DEFINE_CONSTRUCTOR

/****************************************************
	Node building helper routines
*****************************************************/

/**
 * Short cut for creating VCONSTANT node with a given vector value.
 * \ingroup node_constructors
 * \param vec vector to copy 
 * \return pointer to subtree
 */
inline CNode*	cVECTOR( CVector& vec ) 
{
	CVector* v = CVector::AllocFromHeap( CNode::CurrentHeap(), vec.GetWidth() );
	CNode* n;
	*v = vec;
	v->SetPreferredBase( vec.GetPreferredBase() );
	n = new(CNode::CurrentHeap()) CNode( NULL, eVCONSTANT );
	n->Arg<CVector*>(0) = v;
	return n;
}

/**
 * Short cut for creating VCONSTANT node with a given string value.
 * \ingroup node_constructors
 * \param s string to copy 
 * \return pointer to subtree
 */
inline CNode*	cSTRING( const char* s ) 
{
        int len = strlen( s );
        if( !len ) {
             len = 1;
        }
        CVector* v = CVector::AllocFromHeap( CNode::CurrentHeap(), len*8 );
        v->LoadString( s );
	CNode* n = new(CNode::CurrentHeap()) CNode( NULL, eVCONSTANT );
	n->Arg<CVector*>(0) = v;
	return n;
}

/**
 * Short cut for creating VCONSTANT node with a given integer value.
 * \ingroup node_constructors
 * \param i integer to copy 
 * \return pointer to subtree
 */
inline CNode*	cINT32( INT32 i ) 
{
	CVector* v = CVector::AllocFromHeap( CNode::CurrentHeap(), 32 );
	CNode* n;
	*v = i;
	v->Sized(FALSE);
	v->Signed(TRUE);
	v->Based(FALSE);
	n = new(CNode::CurrentHeap()) CNode( NULL, eVCONSTANT );
	n->Arg<CVector*>(0) = v;
	return n;
}

/**
 * Short cut for creating RCONSTANT node with a given double value.
 * \ingroup node_constructors
 * \param number double to copy 
 * \return pointer to subtree
 */
inline CNode* cREAL( double number ) 
{
    CNode* node = new(CNode::CurrentHeap()) CNode( NULL, eRCONSTANT );
    node->Arg<char*>(0) = d2s(number,CNode::CurrentHeap());
    return node;
}

/** \ingroup node_constructors
 * Link together two nodes with an ELIST operator. If one is NULL 
 * do not create a new node, but return the non-null argument. 
 * \param n1 subtree 1
 * \param n2 subtree 2
 * \return pointer to combined subtree
 */
inline CNode*	cELINK( CNode* n1, CNode* n2 )
{
	if( n1 == NULL ) {
		return n2;
	} else if( n2 == NULL ) {
		return n1;
	}
	return cELIST( n1, n2 );
}

/**
 * Short cut for creating an subtree that calculates the absolute
 * value of an expression.
 * \ingroup node_constructors
 * \param a expression
 * \return pointer to absolute value subtree
 */
inline CNode*	cABS( CNode* a ) 
{
	CNode* a1 = a->Clone();
	CNode* a2 = a->Clone();
	CNode* c = cGE(a,cINT32(0));
	return cHOOK( c, a1, cNEG( a2) );
}

/**
 * Short cut for creating an subtree that calculates the absolute
 * difference between two expressions.
 * \ingroup node_constructors
 * \param a expression 1
 * \param b expression 2
 * \return pointer to absolute value subtree
 */
inline CNode*	cABSDIFF( CNode* a, CNode* b ) 
{
	return cABS( cSUB( a, b ) );
}

/**
 * Short cut for linking together to nodes with a LIST operator.
 * A new node is created only if both operands are not NULL.
 * \ingroup node_constructors
 * \param n1 expression 1
 * \param n2 expression 2
 * \return pointer to resultant subtree
 */
inline CNode*	cLINK( CNode* n1, CNode* n2 )
{
	if( n1 == NULL ) {
		return n2;
	} else if( n2 == NULL ) {
		return n1;
	}
	return cLIST( n1, n2 );
}

/**
 * Short cut for creating a expression tree that calculates
 * the maximum of two expressions.
 * \ingroup node_constructors
 * \param n1 expression 1
 * \param n2 expression 1
 * \return pointer to absolute value subtree
 */
inline CNode*	cMAX( CNode* n1, CNode* n2 ) 
{
	CNode* cond = cLT(n2->Clone(),n1->Clone());
	return cHOOK(cond,n1,n2);
}


/****************************************************
	utility routines
*****************************************************/

/**
 * \defgroup node_utilities Node Utilities
 */
/**
 * Walks a list of nodes and collects the specified augments of a given
 * node type. Results are returned as elements of an stl vector.
 * \ingroup node_utilities
 * \param n tree to traverse.
 * \param op operation to search for.
 * \param argNumber argument number to return.
 * \param v containter to return results.
 */
template<class T> void ArgList2Vector(CNode* n, NodeOp_t op, 
				int argNumber, vector<T>& v) 
{
    if( !n ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	ArgList2Vector<T>(n->Arg<CNode*>(0),op,argNumber,v);
	ArgList2Vector<T>(n->Arg<CNode*>(1),op,argNumber,v);
	break;
    default:
	if( n->GetOp() == op ) {
	    v.push_back(n->Arg<T>(argNumber));
	}
	break;
    }
}

/**
 * Walks an expression elist of nodes and collects the subtrees that don't match
 * the given node types. Results are returned as elements of an stl vector.
 * \ingroup node_utilities
 * \param n tree to traverse.
 * \param excludeOps operations to exclude from search
 * \param v containter to return results.
 */
inline void EList2VectorExclude(CNode* n, const set<NodeOp_t>& excludeOps, vector<CNode*>& v) 
{
    if( !n ) {
	return;
    }
    switch( n->GetOp() ) {
    case eELIST:
	EList2VectorExclude(n->Arg<CNode*>(0),excludeOps,v);
	EList2VectorExclude(n->Arg<CNode*>(1),excludeOps,v);
	break;
    default:
	if( excludeOps.find(n->GetOp()) == excludeOps.end() ) {
	    v.push_back(n);
	}
	break;
    }
}

/**
 * Walks an expression list of nodes and collects the subtrees that don't match
 * the given node types. Results are returned as elements of an stl vector.
 * \ingroup node_utilities
 * \param n tree to traverse.
 * \param excludeOps operations to exclude from search
 * \param v containter to return results.
 */
inline void List2VectorExclude(CNode* n, const set<NodeOp_t>& excludeOps, vector<CNode*>& v) 
{
    if( !n ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	List2VectorExclude(n->Arg<CNode*>(0),excludeOps,v);
	List2VectorExclude(n->Arg<CNode*>(1),excludeOps,v);
	break;
    default:
	if( excludeOps.find(n->GetOp()) == excludeOps.end() ) {
	    v.push_back(n);
	}
	break;
    }
}

/**
 * Converts a vector array of CNode* into a linked ELIST of the elements.
 * \ingroup node_utilities
 * \param v vector array to convert.
 * \return linked ELIST of elements
 */
inline CNode* Vector2EList(const vector<CNode*>& v) 
{
    CNode* result = NULL;
    vector<CNode*>::const_reverse_iterator ptr;
    for( ptr = v.rbegin(); ptr != v.rend(); ++ptr ) {
	if( result ) {
	    result = cELIST(*ptr, result);
	} else {
	    result = *ptr;
	}
    }
    return result;
}

/**
 * Converts a list of CNode* into a linked ELIST of the elements.
 * \ingroup node_utilities
 * \param v list to convert.
 * \return linked ELIST of elements
 */
inline CNode* List2EList(list<CNode*>& v) 
{
    CNode* result = NULL;
    list<CNode*>::reverse_iterator ptr;
    for( ptr = v.rbegin(); ptr != v.rend(); ++ptr ) {
	if( result ) {
	    result = cELIST(*ptr, result);
	} else {
	    result = *ptr;
	}
    }
    return result;
}

/**
 * Walks a list/elist of nodes and counts the number of node with the specified
 * operation.
 * \ingroup node_utilities
 * \param n tree to traverse.
 * \param op operation to search for.
 * \return number of nodes matching op.
 */
inline int ListCount(CNode* n, NodeOp_t op) 
{
    int result = 0;
    if( !n ) {
	return result;
    }
    switch( n->GetOp() ) {
    case eLIST:
    case eELIST:
	result += ListCount(n->Arg<CNode*>(0),op);
	result += ListCount(n->Arg<CNode*>(1),op);
	break;
    default:
	if( n->GetOp() == op ) {
	    result = 1;
	}
	break;
    }
    return result;
}

/**
 * Walks a list/elist of nodes and counts the number of non-list nodes
 * \ingroup node_utilities
 * \param n tree to traverse.
 * \return number of nodes matching op.
 */
inline int ListCount(CNode* n) 
{
    int result = 0;
    if( !n ) {
	return result;
    }
    switch( n->GetOp() ) {
    case eLIST:
    case eELIST:
	result += ListCount(n->Arg<CNode*>(0));
	result += ListCount(n->Arg<CNode*>(1));
	break;
    default:
	result = 1;
	break;
    }
    return result;
}

/**
 * \defgroup utilities Utility Routines
 */

/**
 * Convert char string to double.
 * \ingroup utilities
 * \param s char pointer to string to convert.
 * \return converted real.
 */
inline double s2d( char* s ) {
	return atof(s);
}

/**
 * Convert double to char string allocating
 * storage on given heap.
 * \ingroup utilities
 * \param d double to convert.
 * \param heap heap to use for allocation.
 * \return pointer to char string.
 */
inline char* d2s( double d, CObstack* heap ) {
     char buffer[256];
     // note this isn't quite correct as it will turn
     // reals into ints, ie 2.0 => 2
     snprintf( buffer, sizeof(buffer), "%g", d );
     char* s = (char*)heap->Alloc(strlen(buffer)+1);
     strcpy( s, buffer );
     return s;
}

/*********************************************
 * Adjust nodes structure to be efficient for
 * tail recursion
 *********************************************/
inline CNode* RebalanceRight( CNode* n ) {
	if( n == NULL ) {
		return n;
	}
	if( n->GetOp() != eLIST ) {
		return n;
	}
	CNode* result = n;
	CNode* parent = NULL;	
	while( 1 ) {
    	    while( n->Arg<CNode*>(0) && n->Arg<CNode*>(0)->GetOp() == eLIST ) {
    	        CNode* l = n->Arg<CNode*>(0);
    	        CNode* ll = l->Arg<CNode*>(0);
    	        CNode* lr = l->Arg<CNode*>(1);
    	        l->Arg<CNode*>(1) = n;
    	        n->Arg<CNode*>(0) = lr;
    	        n = l;
		if( parent ) { 
		    parent->Arg<CNode*>(1) = n;
		} else {
    	            result = n;
		}
    	    }
    	    if( n->Arg<CNode*>(1) && n->Arg<CNode*>(1)->GetOp() == eLIST ) {
		parent = n;
    	        n = n->Arg<CNode*>(1);
    	    } else {
		break;
	    }
	}
	return result;
}
/*********************************************
 * Analyse tree struct.
 * - depth is worst case stack depth right side
 *   optimization.
 *   count is number nodes.
 *********************************************/
inline void MeasureDepth( CNode* n, int* count, int* depth )
{
    *count = 0;
    *depth = 0;
    if( !n ) {
        return;
    }
    if( n->GetOp() == eLIST ) {
        int count0 = 0;
        int depth0 = 0;
        int count1 = 0;
        int depth1 = 0;
        if( n->Arg<CNode*>(0) ) {
            MeasureDepth( n->Arg<CNode*>(0), &count0, &depth0 );
            depth0++;
        }
        if( n->Arg<CNode*>(1)  ) {
            MeasureDepth( n->Arg<CNode*>(1), &count1, &depth1 );
        }
        *count = count0+count1;
        *depth = depth0 > depth1 ? depth0 : depth1;
    }
    (*count)++;
}


/**************************************************
 * Helper routine for parser to build trees that 
 * are effient for tail recursion.
 *************************************************/
inline CNode_pr	cLINK( CNode_pr pr1, CNode* n2 )
{
    if( !n2 ) {
        return pr1;
    } else if( !pr1.tail ) {
	CNode_pr pr;
	pr.head = n2;
	pr.tail = n2;
	return pr;
    } else if( pr1.tail->GetOp() != eLIST ) {
	CNode* t = cLINK( pr1.head, n2 );
	CNode_pr pr;
	pr.head = t;
	pr.tail = t;
	return pr;
    } else {
        pr1.tail->Arg<CNode*>(1) = cLINK(pr1.tail->Arg<CNode*>(1),n2);
	CNode_pr pr;
	pr.head = pr1.head;
	pr.tail = pr1.tail->Arg<CNode*>(1);
	return pr;
    }
}

/**************************************************
 * convert hierarchical reference tree to string
 *************************************************/
inline string HierarchicalReference2String( CNode* ref )
{
    string buffer;
    switch( ref->GetOp() ) {
    case eARRAY: {
        buffer = HierarchicalReference2String(ref->Arg<CNode*>(0)).c_str();
        vector<CNode*> indexes;
        EList2VectorExclude( ref->Arg<CNode*>(1), 
    			 set<NodeOp_t>(), indexes );
        vector<CNode*>::iterator ptr;
        for( ptr = indexes.begin(); ptr != indexes.end(); ++ptr ) {
            switch( ref->Arg<CNode*>(0)->GetOp() ) {
            case eSLICE:
            case ePSLICE:
            case eMSLICE:
    	    break;
            default: {
                INT32 value = (*ptr)->EvalINT32();
    	    ostringstream subscript;
    	    subscript << '[' << value << ']'; 
    	    buffer += subscript.str();
    	    } break;
            }
        }
        } break;
    case eMEMBER:
        buffer = HierarchicalReference2String(ref->Arg<CNode*>(0)).c_str();
        buffer += ".";
        buffer += ref->Arg<CSymbol*>(1)->GetName();
        break;
    case eEXTERNAL_REF:
        buffer = ref->Arg<CSymbol*>(0)->GetName();
        ref = NULL;
        break;
    case eNET_REF:		
    case eVAR_REF:		
    case ePARAM_REF:		
    case ePORT_REF:		
    case eFWD_REF:		
    case eGENVAR_REF:		
        buffer = ref->Arg<CDecl*>(0)->GetName();
        ref = NULL;
        break;
    default:;
        MASSERT( FALSE );
    }
    return buffer;
}

/*****************************************************
 * Support routines for cnode_def.h constructs
 *****************************************************/
inline CNode* cMAX_N( CNode* first, ... )
{
    CNode* result = first;
    va_list ap;
    va_start( ap, first );
    while( 1 ) {
        CNode* arg = va_arg(ap,CNode*);
	if( !arg ) {
	    break;
        }
	if( !Equivalent( result, arg ) ) {
            result = cMAX( result, arg );
	}
    }
    va_end( ap );
    return result;
}

inline CNode* cADD_N( CNode* first, ... )
{
    CNode* result = first;
    va_list ap;
    va_start( ap, first );
    while( 1 ) {
        CNode* arg = va_arg(ap,CNode*);
	if( !arg ) {
	    break;
        }
	result = cADD( result, arg );
    }
    va_end( ap );
    return result;
}

inline CNode* cMUL_N( CNode* first, ... )
{
    CNode* result = first;
    va_list ap;
    va_start( ap, first );
    while( 1 ) {
        CNode* arg = va_arg(ap,CNode*);
	if( !arg ) {
	    break;
        }
	result = cMUL( result, arg );
    }
    va_end( ap );
    return result;
}

inline CNode* cABSDIFFPLUS1_N( CNode* first, ... )
{
    va_list ap;
    va_start( ap, first );
    CNode* second = va_arg(ap,CNode*);
    /*
     * only make sense for 2 args
     */
    MASSERT( va_arg(ap,CNode*) == NULL );
    va_end( ap );
//    return cADD(cABSDIFF(first,second),cINT32(1));
    return cHOOK(
		cGE(first,second),
		cADD(cSUB(first->Clone(),second->Clone()),cINT32(1)),
		cADD(cSUB(second->Clone(),first->Clone()),cINT32(1)) );
		
}

inline int cMAX( int a1, int a2 )
{
    return a1 < a2 ? a2 : a1;
}

inline int cMAX( int a1, int a2, int a3 )
{
    return cMAX(a1,cMAX(a2,a3));
}

inline int cADD( int a1, int a2 )
{
    return a1 + a2;
}

inline int cMUL( int a1, int a2 )
{
    return a1 * a2;
}

inline int cABSDIFFPLUS1( int a1, int a2 )
{
    int diff = a1-a2;
    return (diff < 0 ? -diff : diff)+1;
}


#endif // CNODE_HPP

