/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cvector.hpp
 *		- abstract class for bit vector declaration 
 *
 *
 ******************************************************************************
 */

#ifndef CVECTOR_HPP
#define CVECTOR_HPP

#include <stdio.h>
#include <math.h>
#include "glue.h"
#include "cuint.h"
#include "cdecl.h"


/**
 * Bit vector class for implementing 4 state verilog
 * signed and unsigned arithmetic. bits specified by '?'
 * are maintained in representation. Information is also
 * maintained about the constant specification format so
 * it can be faithfully be reproduce later. The class maintains
 * a single char buffer for converions to character strings.
 * The result must be immediatedly be copied otherwise the 
 * buffer will be overwritten on the next conversion.
 *
 * Bit representation:
 * 	- aval=0, bval=0, hook=* => 0
 * 	- aval=1, bval=0, hook=* => 1
 * 	- aval=0, bval=1, hook=0 => z
 * 	- aval=0, bval=1, hook=1 => ?
 * 	- aval=1, bval=1, hook=* => x
 */
class CVector
{
private:
	CUInt		aval;	       ///< aval part of vector
	CUInt		bval;	       ///< bval part of vector
	CUInt		hook;	       ///< specifies '?' (aval=0,bval=1)
	INT32		width;	       ///< number of bits in vector.
	int		sized;	       ///< non-zero if constant was specifed
				       ///< with a size specification
	int		based;	       ///< non-zero if constant was specified
				       ///< with a base specification.
	int		unbased;       ///< non-zero if constant was specified
				       ///< with a unbase specification.
	int		preferredBase; ///< default output base: 
				       ///< 2,8,10,16, or 0 for string
	int		_signed;       ///< non-zero if constant is signed
	int		overflowed; // set to 1 if string conversion overflowed

	static int	bufferSize;    ///< size in bytes of conversion buffer.
	static char*	buffer;	       ///< ptr to conversion buffer.
public:
	/**
 	 * Create vector allocating all storage from given heap.
 	 * \param aHeap obstack to allocate storage from.
 	 * \param width width of vector in bits.
 	 * \return pointer to vector object.
 	 */
	static CVector* AllocFromHeap( CObstack* aHeap, int width ) {
		CVector* v = new(aHeap) CVector( width );
		v->SetHeap( aHeap );	
		return v;
	}

	/**
 	 * Create bit vector of given width.
 	 * \param aWidth width of vector in bits.
 	 */
	CVector( INT32 aWidth ); 
	/**
 	 * Destroy vector freeing storage if possible.
 	 */
	~CVector();
	/**
 	 * Calculate a hash for vector value. 
 	 * If a and b are vectors and the value of a == b then
 	 * a.Hash() == b.Hash()
 	 * \return hash of vector value.
 	 */
	unsigned Hash();
	/**
 	 * Determine if any bit is either x or z.
 	 * \return non-zero if any bit is either z or x.
 	 */
	int	HasXZ() { return bval != 0; }
	/**
 	 * Determine if any bit is a z.
 	 * \return non-zero if any bit is a z.
 	 */
	int	HasZ( void ) { return (~aval & bval) != 0; }
	/**
 	 * Determine if value is negative.
 	 * \return non-zero if value is signed and msb is 1.
 	 */
	int	IsNegative() 
		{ return _signed && ((aval>>(width-1)) & 1) != 0 && !HasXZ(); }
	/**
 	 * Set preferred base for printing value.
 	 * \param base base for printing: 2, 8, 10, 16 or 0 for string.
 	 */
	void	SetPreferredBase( int base ) { preferredBase = base; }
	/**
 	 * Get preferred base for printing value.
 	 * \return base base for printing: 2, 8, 10, 16 or 0 for string.
 	 */
	int	GetPreferredBase() { return preferredBase; }
	/**
 	 * Get sized attribute.
 	 * \return non-zero if constant was specified with a size
 	 * 	   specification.
 	 */
	int	Sized() { return sized; }
	/**
 	 * Set sized attribute.
 	 * \param sized non-zero if constant was specified with a size
 	 * 	   specification.
 	 */
	void	Sized( int sized ) { this->sized = sized; }
	/**
 	 * Get based attribute.
 	 * \return non-zero if constant was specified with a base
 	 * 	   specification.
 	 */
	int	Based() { return based; }
	/**
 	 * Set based attribute.
 	 * \param based non-zero if constant was specified with a base
 	 * 	   specification.
 	 */
	void	Based( int based ) { this->based = based; }
	/**
 	 * Get unbased attribute.
 	 * \return non-zero if constant was specified with a unbase
 	 * 	   specification.
 	 */
	int	Unbased() { return unbased; }
	/**
 	 * Set unbased attribute.
 	 * \param unbased non-zero if constant was specified with a unbased
 	 * 	   specification.
 	 */
	void	Unbased( int unbased ) { this->unbased = unbased; }
	/**
 	 * Get signed attribute.
 	 * \return non-zero if constant is signed.
 	 */
	int	Signed() const { return _signed; }
	/**
 	 * set signed attribute.
 	 * \param _signed non-zero if constant is signed.
 	 */
	void	Signed( int _signed ) { this->_signed = _signed; }
	/**
 	 * Determine if conversion from ascii overflowed the
 	 * give vector width.
 	 * \return non-zero if conversion overflowed.
 	 */ 
	int	Overflowed() { return overflowed; }
	/**
 	 * Get vector bit width.
 	 * \return width of vector in bits.
 	 */
	INT32	GetWidth( void );
	/**
 	 * Set width of vector in bits.
 	 * If new width is smaller then current width, value is truncated.
 	 * If new width is larger and vector is signed, value is sign
 	 * extended.
 	 * \param newWidth new vector width.
 	 */
	void	SetWidth( INT32 newWidth );
	/**
 	 * Set vector value to X
 	 */
	void	SetToX();
	/**
 	 * Assignment operator.
 	 * \param v vector to assign from.
 	 */
	const CVector& operator=( const CVector& v );
	/**
 	 * Assignment operator.
 	 * \param v integer to assign to vector.
 	 */
	UINT32   operator=( UINT32 v );
	/**
 	 * Load decimal value from string.
 	 * \param string string to convert.
 	 */
	int	LoadDecimal( const char* string );	
	/**
 	 * Load binary value from string.
 	 * \param string string to convert.
 	 */
	int	LoadBinary( const char* string );	
	/**
 	 * Load octal value from string.
 	 * \param string string to convert.
 	 */
	int	LoadOctal( const char* string );	
	/**
 	 * Load hex value from string.
 	 * \param string string to convert.
 	 */
	int	LoadHex( const char* string );	
	/**
 	 * Load string value from string.
 	 * \param string string to convert.
 	 */
	int	LoadString( const char* string );	
	/**
 	 * Convert vector to string using preferred base.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char* 	GetVString( void );
	/**
 	 * Convert vector to decimal string.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char*	GetDecimal( void );	
	/**
 	 * Convert vector to binary string.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char*	GetBinary( void );	
	/**
 	 * Convert vector to octal string.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char*	GetOctal( void );	
	/**
 	 * Convert vector to hex string.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char*	GetHex( void );	
	/**
 	 * Convert vector to ascii string. 
 	 * Conversion includes leading and trailing quote.
 	 * \return buffer containing converted value. Buffer is
 	 * 	   class static storage and should be used or copied
 	 * 	   before next vector conversion.
 	 */
	char*	GetString( void );	
	/**
 	 * Equality operator
 	 * \param v vector to compare.
 	 * \return non-zero if vectors are equal.
 	 */
	int operator==( CVector& v );
	/**
 	 * Equality operator
 	 * \param i integer to compare.
 	 * \return non-zero if integer equals this vector
 	 */
	int operator==( UINT32 i );
	/**
 	 * Inequality operator
 	 * \param v vector to compare.
 	 * \return non-zero if vectors are not equal.
 	 */
	int operator!=( CVector& v ) { return !(*this == v); }
	/**
 	 * Inequality operator
 	 * \param i integer to compare.
 	 * \return non-zero if integer is not equal to this vector
 	 */
	int operator!=( UINT32 i ) { return !(*this == i); }
	/**
 	 * Get value as a 32 bit integer. Value is truncated
 	 * if it is larger than 32 bits.
 	 * \return signed 32 bit integer value.
 	 */
	INT32	GetINT32( void )
	{
		if( IsNegative() ) {
		    CVector n(GetWidth());
		    n.Signed(1);
		    Neg( &n, this );
		    return -(INT32)n.aval.GetUINT32();
		}
		
		return aval.GetUINT32();
	}
	/**
 	 * Get value as a 64 bit integer. Value is truncated
 	 * if it is larger than 64 bits.
 	 * \return signed 64 bit integer value.
 	 */
	INT64	GetINT64( void )
	{
		if( IsNegative() ) {
		    CVector n(GetWidth());
		    n.Signed(1);
		    Neg( &n, this );
		    return -(INT64)n.aval.GetUINT64();
		}
		
		return aval.GetUINT64();
	}
	/**
 	 * Load vector with integer value.
 	 * \param v integer to load.
 	 */
	void	LoadINT32( INT32 v )
	{
		aval = (unsigned int)v;	
		bval = 0;
	}
	/**
 	 * Load vector with integer part of real value.
 	 * \param d real value to load.
 	 */
        void LoadReal( double d );
	/**
 	 * Get vector value as a real.
 	 * \return real value.
 	 */
        double GetReal();
	/**
 	 * Set vector to aval part of this vector.
 	 * \param v vector to load with aval.
 	 */
	void	GetAval( CVector& v ) {
		v.aval = aval;
		bval = 0;
	}
	/**
 	 * Set vector to bval part of this vector.
 	 * \param v vector to load with bval.
 	 */
	void	GetBval( CVector& v ) {
		v.aval = bval;
		bval = 0;
	}

	/**
 	 * \cond
 	 */
	friend void	Add( CVector* r, CVector* a, CVector* b );
	friend void	Sub( CVector* r, CVector* a, CVector* b );
	friend void	Mul( CVector* r, CVector* a, CVector* b );
	friend void	Div( CVector* r, CVector* a, CVector* b );
	friend void	Pow( CVector* r, CVector* a, CVector* b );
	friend void	Lsha( CVector* r, CVector* a, CVector* b );
	friend void	Rsha( CVector* r, CVector* a, CVector* b );
	friend void	Lsh( CVector* r, CVector* a, CVector* b );
	friend void	Rsh( CVector* r, CVector* a, CVector* b );
	friend void	Cat( CVector* r, CVector* a, CVector* b );
	friend void	Rep( CVector* r, CVector* a, CVector* b );
	friend void	Mod( CVector* r, CVector* a, CVector* b );
	friend void	And( CVector* r, CVector* a, CVector* b );
	friend void	Or( CVector* r, CVector* a, CVector* b );
	friend void	Xor( CVector* r, CVector* a, CVector* b );
	friend void	Xnor( CVector* r, CVector* a, CVector* b );
	friend void	Com( CVector* r, CVector* a );
	friend void	Neg( CVector* r, CVector* a );
	friend void	Plus( CVector* r, CVector* a );
	friend void	Not( CVector* r, CVector* a );
	friend void	Le( CVector* r, CVector* a, CVector* b );
	friend void	Lt( CVector* r, CVector* a, CVector* b );
	friend void	Ge( CVector* r, CVector* a, CVector* b );
	friend void	Gt( CVector* r, CVector* a, CVector* b );
	friend void	Eq( CVector* r, CVector* a, CVector* b );
	friend void	Ne( CVector* r, CVector* a, CVector* b );
	friend void	Ceq( CVector* r, CVector* a, CVector* b );
	friend void	Cne( CVector* r, CVector* a, CVector* b );
	friend void	Lor( CVector* r, CVector* a, CVector* b );
	friend void	Land( CVector* r, CVector* a, CVector* b );
	friend void	Rand( CVector* r, CVector* a );
	friend void	Ror( CVector* r, CVector* a );
	friend void	Rnand( CVector* r, CVector* a );
	friend void	Rnor( CVector* r, CVector* a );
	friend void	Rxor( CVector* r, CVector* a );
	friend void	Rxnor( CVector* r, CVector* a );
	friend void	Hook( CVector* r, CVector* c, CVector* a, CVector* b );
	/**
 	 * These routines should not be used by plugins.
 	 */
	int IsWidthConstant() { return TRUE; }
	int IsWidthVolatile() { return FALSE; }
	int IsWidthEvaluateable() { return TRUE; }
	CNode* GetWidthExp();
	virtual NodeType_t GetNodeType( void );
private:
	void Adjust( void );
	void GrowBuffer( int bytes )
	{
		if( bytes <= bufferSize ) {
			return;
		}
		if( buffer != NULL ) {
			shell_xfree( buffer, bufferSize );
		}
		bufferSize = bytes * 2;
		buffer = (char*) shell_xmalloc( bufferSize );
	}
	void SetHeap( CObstack* aHeap ) {
		aval.SetHeap( aHeap );
		bval.SetHeap( aHeap );
	}
	/**
 	 * Allocate base object from obstack.
 	 * \param size size of object in bytes.
 	 * \param pointer to obstack to use for allocation.
 	 * \return pointer to allocated storage.
 	 */
        void* operator new(size_t size, CObstack* heap) {
                return heap->Alloc( size );
        }       
	/**
 	 * \endcond
 	 */
};

/**
 * These routines are used by the expression evaluation and should not be used
 * by plugins.
 * \cond
 */
inline void Le( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a <= *b );
}


inline void Le( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a <= b );
}

inline void Lt( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a < *b );
}


inline void Lt( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a < b );
}

inline void Ge( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a >= *b );
}


inline void Ge( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a >= b );
}

inline void Gt( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a > *b );
}


inline void Gt( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a > b );
}


inline void Ne( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a != *b );
}

inline void Ne( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a != b );
}

inline void Eq( CVector* r, double* a, double* b )
{
	r->LoadINT32( *a == *b );
}

inline void Eq( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a == b );
}

inline void Cne( CVector*, double*, double* )
{
	fatal( NULL, "!== illegal for reals" );
}

inline void Cne( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a != b );
}

inline void Ceq( CVector*, double*, double* )
{
	fatal( NULL, "=== illegal for reals" );
}

inline void Ceq( CVector* r, INT32 a, INT32 b )
{
	r->LoadINT32( a == b );
}

/**
 *Compute logical value of expression
 */
inline void	LogicalValue( CVector* dst, CVector* a ) 
{
	CVector zero(a->GetWidth());
	Ne( dst, a, &zero );
}

inline void Hook( double* r, CVector* c, double* a, double* b )
{
	CVector cond(1);
	LogicalValue( &cond, c );
	if( cond.HasXZ() ) {
	    *r = 0;
	} else if( cond == 0) {
		*r = *b;
	} else {
		*r = *a;
	}	
}

inline void Cat( double*, double*, double* )
{
	fatal( NULL, "=== illegal for reals" );
}
/**
 * \endcond
 */

#endif // CVECTOR_HPP
